/*  Copyright (C) 2003-2007  CAMP
 *  Copyright (C) 2005       CSC - IT Center for Science Ltd.
 *  Please see the accompanying LICENSE file for further information. */

#include "fd.c"
#include "wfd.c"
#include "relax.c"
#include "wrelax.c"
#include "cut.c"
#include "zero.c"
#include "paste.c"
#include "spline.c"
#include "stencils.c"
#include "restrict.c"
#include "translate.c"
#include "interpolate.c"

#define BMGSCOMPLEX
#include "fd.c"
#include "wfd.c"
#include "cut.c"
#include "zero.c"
#include "paste.c"
#include "restrict.c"
#include "interpolate.c"
